// SNF4CGP/JobPool.cpp
// Copyright (C) 2009 ARM Research Labs, LLC.
// See www.armresearch.com for more information.

#include "JobPool.hpp"
#include "Command.hpp"
#include "ScannerPool.hpp"
#include "OutputProcessor.hpp"

#include "../SNFMulti/SNFMulti.hpp"

#include "../CodeDweller/threading.hpp"
#include "../CodeDweller/faults.hpp"

#include <iostream>
#include <string>
#include <vector>

using namespace std;

//// Tuning Constants //////////////////////////////////////////////////////////

const int ReadBufferSize = snf_ScanHorizon;
const int StringReserveSize = 2048;

//// Job ///////////////////////////////////////////////////////////////////////

void Job::emitCommentPrefix() {
}

void Job::emitComment(const string& Comment) {
    ostringstream O;
    O << "* " << Comment << endl;
    OutputBuffer.append(O.str());
}

void Job::emitOK() {
    ostringstream O;
    O << CurrentCommand.Number << " OK" << endl;
    OutputBuffer.append(O.str());
}

const string INTFVersion = "1";

void Job::emitINTF() {
    ostringstream O;
    O << CurrentCommand.Number << " INTF " << INTFVersion << endl;
    OutputBuffer.append(O.str());
}

void Job::emitFAILURE() {
    ostringstream O;
    O << CurrentCommand.Number << " FAILURE" << endl;
    OutputBuffer.append(O.str());
}

string formatAsCGPString(const string& S) {
    string CGPString;
    CGPString.push_back('"');
    for(size_t i = 0; i < S.length(); i++) {
        char C = S[i];
        switch(C) {
            case '\r': break;
            case '\t': { CGPString.append("\\t"); break; }
            case '\n': { CGPString.append("\\e"); break; }
            case '\\': { CGPString.append("\\"); break; }
            case '"': { CGPString.append("\\\""); break; }
            default: CGPString.push_back(C);
        }
    }
    CGPString.push_back('"');
    return CGPString;
}

void Job::emitADDHEADER(const string& Headers) {
    ostringstream O;
    O << CurrentCommand.Number << " ADDHEADER " << formatAsCGPString(Headers) << endl;
    OutputBuffer.append(O.str());
}

void Job::emitERROR(const string& Report) {
    ostringstream O;
    O << CurrentCommand.Number << " ERROR " << formatAsCGPString(Report) << endl;
    OutputBuffer.append(O.str());
}

void Job::emitDISCARD() {
    ostringstream O;
    O << CurrentCommand.Number << " DISCARD" << endl;
    OutputBuffer.append(O.str());
}

void Job::emitREJECTED(const string& Report) {
    ostringstream O;
    O << CurrentCommand.Number << " REJECTED " << formatAsCGPString(Report) << endl;
}

void Job::finalize() {
    Output.outputJob(*this);
}

void Job::doWakeUp() {
    emitComment("SNF4CGP Waking Up");
    emitComment(CurrentCommand.Data);
}

void Job::doINTF() {
    emitINTF();
}

void Job::doFAIL() {
    ostringstream O;
    O << "* SNF4CGP[" << CurrentCommand.Number << "] Does not understand: "
      << formatAsCGPString(CurrentCommand.Data) << endl;
    OutputBuffer.append(O.str());
    emitFAILURE();
}

void Job::doFILE() {
    doRead();
    doScan();
    doAction();
    closeReader();
}

RuntimeCheck CheckMessageReaderIsValid("Job::Reader() Check(0 != Reader_)");

ifstream& Job::Reader() {                                                       // Safe access to Reader_.
    CheckMessageReaderIsValid(0 != Reader_);
    return (*Reader_);
}

LogicFault FaultMessageReaderOpenedTwice("Job::openReader() Fault(0 != Reader_)");
RuntimeCheck CheckopenReaderWasSuccessful("Job::openReader() Check(Reader().good())");

void Job::openReader(string Path) {
    FaultMessageReaderOpenedTwice(0 != Reader_);
    Reader_ = new ifstream(Path.c_str(), ios::binary);
    CheckopenReaderWasSuccessful(Reader().good());
}

void Job::closeReader() {
    if(Reader_) {
        Reader().close();
        delete Reader_;
        Reader_ = 0;
    }
}

RuntimeCheck CheckJobWriterIsValid("Job::Writer() Check(0 != Writer_)");

ofstream& Job::Writer() {                                                       // Safe access to Writer_.
    CheckJobWriterIsValid(0 != Writer_);
    return (*Writer_);
}

LogicFault FaultMessageWriterOpenedTwice("Job::openWriter() Fault(0 != Writer_)");
RuntimeCheck CheckopenWriterWasSuccessful("Job::openWriter() Check(Writer().good())");

void Job::openWriter(string Path) {
    FaultMessageWriterOpenedTwice(0 != Writer_);
    Writer_ = new ofstream(Path.c_str(), ios::trunc | ios::binary);
    CheckopenWriterWasSuccessful(Writer().good());
}

void Job::closeWriter() {
    if(Writer_) {
        Writer().close();
        delete Writer_;
        Writer_ = 0;
    }
}

void Job::doRead() {
    openReader(Job::CurrentCommand.Data);
    ReadBuffer.assign(ReadBufferSize, 0);
    Reader().read(reinterpret_cast<char*>(&ReadBuffer[0]), ReadBuffer.size());
}

string Job::ScanName() {                                                        // Scan name from FILE command
    ostringstream ScanNameFormatter;
    ScanNameFormatter << "[" << CurrentCommand.Number << "]" << CurrentCommand.Data;
    return ScanNameFormatter.str();
}

void Job::setConfigurationFromScanResult(ScopeScanner& myScanner) {
    ScanResultConfiguration = Scanners.ConfigurationForResultCode(ScanResultCode);
    if(ScanResultConfiguration.InjectHeaders) HeadersToInject = myScanner.Engine().getXHDRs();
    if(ScanResultConfiguration.EmitXMLLog) XMLLogData = myScanner.Engine().getXMLLog();
    if(ScanResultConfiguration.EmitClassicLog) ClassicLogData = myScanner.Engine().getClassicLog();
}

void Job::doScan() {
    ScopeScanner myScanner(Scanners);
    ScanResultCode =
      myScanner.Engine().scanMessage(
        &ReadBuffer[0], Reader().gcount(), ScanName(), JobTimer.getElapsedTime()
    );
    setConfigurationFromScanResult(myScanner);
}

void Job::doAction() {
    // Select action based on current configuration
    // Do the appropriate action
}

void Job::doBypass() {
}

void Job::doAllow() {
}

void Job::doReject() {
}

void Job::doDelete() {
}

void Job::emitMessageMovedTo(string& Destination) {
    ostringstream O;
    O << "* SNF4CGP[" << CurrentCommand.Number << "] Moved message to " << Destination << endl;
    OutputBuffer.append(O.str());
}

void Job::identifyLocalLineEnd() {                                              // Identify the line endings in use:
    if(0 < LocalLineEnd.length()) return;                                       // Do this only once for any system.
    for(int i = 0; i < Reader().gcount(); i++) {                                // Look for the first line end.
        if('\r' == ReadBuffer[i] && '\n' == ReadBuffer[i+1]) {                  // It may be \r\n.
            LocalLineEnd = "\r\n"; return;
        } else
        if('\n' == ReadBuffer[i] && '\n' == ReadBuffer[i+1]) {                  // It may be \n. (find block end \n\n)
            LocalLineEnd = "\n"; return;
        }
    }
}

RuntimeCheck CheckLineEndsIdentified("Job::findHeaderInsertPoint() Check(0 < LocalLineEnd.length())");

size_t Job::findHeaderInsertPoint() {                                           // How to find the insert point:
    identifyLocalLineEnd();
    CheckLineEndsIdentified(0 < LocalLineEnd.length());
    string BlockTerminus = LocalLineEnd + LocalLineEnd;
    size_t InsertCursor = 0;

    //// Search for the end of the CGP control block.

    for(;InsertCursor < (Reader().gcount() - BlockTerminus.length()); InsertCursor++) {
        if(0 == BlockTerminus.compare(reinterpret_cast<char*>(&ReadBuffer[InsertCursor]))) break;
    }

    //// Search for the end of the message headers.

    InsertCursor += BlockTerminus.length();
    for(;InsertCursor < (Reader().gcount() - BlockTerminus.length()); InsertCursor++) {
        if(0 == BlockTerminus.compare(reinterpret_cast<char*>(&ReadBuffer[InsertCursor]))) break;
    }

    //// The correct insert point is between the two LocalLineEnd in the BlockTerminus

    InsertCursor += LocalLineEnd.length();

    return InsertCursor;
}

void Job::writeUpToInjectionPoint(size_t InjectionPoint) {
}

void Job::writeHeaders() {
}

void Job:: writeTheRestOfTheFirstBuffer() {
}

void Job::injectHeadersInMessageMove() {
    //// writeUpToTheInjectionPoint()
    //// writeTheHeaders()
    //// writeTheRestOfTheFirstBuffer()
}

void Job::copyRestOfLongMessage() {
    //// while(MoreToDo) {
    ////     readBufferFull()
    ////     writeBufferFull()
    //// }
}

void Job::moveMessageToHoldPath(string& Destination) {
    openWriter(Destination);
    if(ScanResultConfiguration.InjectHeaders) injectHeadersInMessageMove();
    copyRestOfLongMessage();
    closeWriter();
}

const string PathSeparators = "\\/";

string FileNamePart(string Path) {                                              // How to get the file name part of a path:
    string::size_type Found = Path.find_last_of(PathSeparators);
    if(string::npos != Found) Path = Path.substr(Found+1);                      // Return all after the last separator
    return Path;                                                                // or Path if no separators are found.
}

string Job::calculateDestinationFileName() {                                    // How to get the message move destination:
    string Source = CurrentCommand.Data;
    string Destination = ScanResultConfiguration.HoldPath;
    Destination.append(FileNamePart(Source));
    return Destination;
}

void Job::doHold() {
    string Destination = calculateDestinationFileName();
    moveMessageToHoldPath(Destination);
    emitMessageMovedTo(Destination);
}

const int NoScanYet = -1;

Job::Job(ScannerPool& S, OutputProcessor& O) :                                  // Construct with important links.
  Scanners(S),
  Output(O),
  ScanResultCode(NoScanYet),
  Reader_(0),
  Writer_(0) {                                                                  // Minimize heap thrashing.
    OutputBuffer.reserve(StringReserveSize);
    HeadersToInject.reserve(StringReserveSize);
    XMLLogData.reserve(StringReserveSize);
    ClassicLogData.reserve(StringReserveSize);
    ReadBuffer.reserve(ReadBufferSize);
}

Job::~Job() {                                                                   // Cleanup when destructing.
    try { clear(); }
    catch(...) {}
}

void Job::clear() {                                                             // Cleanup for the next command.
    CurrentCommand.clear();
    ScanResultCode = NoScanYet;
    OutputBuffer.clear();
    HeadersToInject.clear();
    XMLLogData.clear();
    ClassicLogData.clear();
    ReadBuffer.clear();
    closeReader();
    closeWriter();
    JobTimer.clear();
}

LogicFault FaultIfQuitGetsHere("Job::setCommand() Fault(Command::QUIT == C.Type)");

void Job::setCommand(Command& C) {                                              // Assign a command for this job.
    JobTimer.start();
    FaultIfQuitGetsHere(Command::QUIT == C.Type);
    CurrentCommand = C;
}

void Job::executeCommand() {
    switch(CurrentCommand.Type) {
        case Command::WAKE: { doWakeUp(); break; }
        case Command::INTF: { doINTF(); break; }
        case Command::FILE: { doFILE(); break; }
        default: { doFAIL(); break; }
    }
}

void Job::emitException(const string& What) {
    ostringstream O;
    O << "* SNF4CGP[" << CurrentCommand.Number << "] Exception: " << What << endl;
    OutputBuffer.append(O.str());
}

void Job::emitUnknownException() {
    ostringstream O;
    O << "* SNF4CGP[" << CurrentCommand.Number << "] Unknown Exception!" << endl;
    OutputBuffer.append(O.str());
}

void Job::doIt() {                                                              // Get the job done.
    try {                                                                       // emitFAILURE() on all exceptions.
        try { executeCommand(); }                                               // executeCommand() and report exceptions.
        catch(exception& e) { emitException(e.what()); throw; }
        catch(...) { emitUnknownException(); throw; }
    }
    catch(...) { emitFAILURE(); }
    finalize();
}

string& Job::Responses() {                                                      // Read the job's report.
    return OutputBuffer;
}

//// JobPool ///////////////////////////////////////////////////////////////////

JobPool::JobPool() :                                                            // Simple constructor.
  Output_(0),
  Scanners_(new ScannerPool()),
  AllocatedJobs(0),
  Started(false) {
}

JobPool::~JobPool() {                                                           // Clean up on the way out.
    try {
        stop();
        Output_ = 0;
        if(Scanners_) delete Scanners_;
        Scanners_ = 0;
    }
    catch(...) {}
}

RuntimeCheck CheckForValidOutputPool("JobPool::Output() Check(0 != Output_)");

OutputProcessor& JobPool::Output() {
    CheckForValidOutputPool(0 != Output_);
    return (*Output_);
}

RuntimeCheck CheckForValidScannerPool("JobPool::Scanners() Check(0 != Scanners_)");

ScannerPool& JobPool::Scanners() {
    CheckForValidScannerPool(0 != Scanners_);
    return (*Scanners_);
}

// Watch out -- Initializing the JobPool also means starting up SNFMulti and
// the ScannerPool.

void JobPool::init(string Configuration, OutputProcessor& O) {                  // Initialize the JobPool.
    Output_ = &O;
    Scanners().init(Configuration);
    ScopeMutex Busy(AllocationMutex);
    Job* FirstJob = makeJob();
    Jobs.give(FirstJob);
    Started = true;
}

Job* JobPool::makeJob() {                                                       // Create and count a Job.
    Job* NewJob = new Job(Scanners(), Output());
    ++AllocatedJobs;
    return NewJob;
}

LogicCheck CheckGrabJobsOnlyWhenStarted("JobPool::grab() Check(Started)");
RuntimeCheck CheckForValidGrabbedJob("JobPool::grab() Check(0 != GrabbedJob)");

Job& JobPool::grab() {                                                          // Grab a job (prefer from the pool).
    ScopeMutex Busy(AllocationMutex);
    CheckGrabJobsOnlyWhenStarted(Started);
    Job* GrabbedJob = 0;
    if(0 < Jobs.size()) GrabbedJob = Jobs.take();                               // Prefer jobs from the pool.
    else GrabbedJob = makeJob();                                                // Make new ones if needed.
    CheckForValidGrabbedJob(0 != GrabbedJob);
    return (*GrabbedJob);
}

void JobPool::drop(Job& J) {                                                    // Drop a job into the pool.
    Jobs.give(&J);
}

void JobPool::killJobFromPool() {                                               // Kill and count a Job from the pool.
    Job* JobToKill = 0;
    JobToKill = Jobs.take();
    delete JobToKill;
    --AllocatedJobs;
}

// Watch out -- Stopping the JobPool also means shutting down SNFMulti and
// the ScannerPool.

void JobPool::stop() {                                                          // Shut down the JobPool.
    ScopeMutex Busy(AllocationMutex);
    while(0 < AllocatedJobs) killJobFromPool();
    Scanners().stop();
    Started = false;
}

